from types import SimpleNamespace
from cogment.api.common_pb2 import Feedback


class Actor:
    def __init__(self, actor_class, actor_id, id_in_class, trial):
        self.actor_class = actor_class
        self.actor_id = actor_id
        self.id_in_class = id_in_class

        self._feedback = []
        self.trial = trial

    def add_feedback(self, value, confidence, tick_id=None, user_data=None):
        if tick_id is None:
            tick_id = -1

        self._feedback.append((tick_id, value, confidence, user_data))


class Trial:
    def __init__(self, id, settings, actor_counts):
        self.id = id
        self.actor_counts = actor_counts
        self.actors = SimpleNamespace(all=[])
        self.settings = settings
        self.tick_id = 0

        actor_id = 0
        for class_index, actor_class in enumerate(self.settings.actor_classes):
            actor_list = []
            id_in_class = 0
            for i in range(actor_counts[class_index]):
                actor = Actor(actor_class, actor_id, id_in_class, self)
                actor_list.append(actor)
                self.actors.all.append(actor)
                actor_id += 1
                id_in_class += 1

            setattr(self.actors, actor_class.id, actor_list)

    def new_observation_table(self):
        return self.settings.ObservationsTable(self.actor_counts)

    def _get_all_feedback(self):
        for actor in self.actors.all:
            a_fb = actor._feedback
            actor._feedback = []

            for fb in a_fb:
                re = Feedback(
                    actor_id=actor.actor_id,
                    tick_id=fb[0],
                    value=fb[1],
                    confidence=fb[2]
                )
                if fb[3] is not None:
                    re.content = fb[3].SerializeToString()

                yield re



from cogment.api.orchestrator_pb2 import (
    TrialStartRequest, TrialFeedbackRequest, TrialActionRequest,
    TrialEndRequest)

from cogment.api.orchestrator_pb2_grpc import TrialStub
from cogment.api.common_pb2 import Action

from cogment.delta_encoding import DecodeObservationData

from cogment.trial import Trial
import grpc


class ClientTrial(Trial):
    def __init__(self, conn, trial_start_rep, settings, actor_class,
                 actor_counts, initial_observation):
        super().__init__(trial_start_rep.trial_id, settings, actor_counts)
        self.connection = conn
        self.observation = initial_observation
        self.actor_id = trial_start_rep.actor_id
        self.actor_class = actor_class
        self.latest_reward = None

    # Perform an action on the trial, and advance time
    def do_action(self, action):
        self.flush_feedback()

        # Send the update to the orchestrator
        update = self.connection.stub.Action(TrialActionRequest(
            trial_id=self.id,
            actor_id=self.actor_id,
            action=Action(content=action.SerializeToString())))

        self.observation = DecodeObservationData(
            self.actor_class,
            update.observation.data,
            self.observation)

        self.latest_reward = update.reward
        self.tick_id = update.observation.tick_id

        # Return the latest observation
        return self.observation, self.latest_reward

    # Kill the trial
    def end(self):
        self.flush_feedback()
        self.connection.stub.End(TrialEndRequest(trial_id=self.id))

    def flush_feedback(self):
        feedbacks = list(self._get_all_feedback())

        if feedbacks:
            req = TrialFeedbackRequest(trial_id=self.id)

            req.feedbacks.extend(feedbacks)
            self.connection.stub.GiveFeedback(req)


class _Connection_impl:
    def __init__(self, stub, settings):
        if not settings:
            raise Exception("missing settings")

        if not stub:
            raise Exception("missing grpc connection stub")

        self.stub = stub
        self.settings = settings

    def start_trial(self, actor_class, trial_cfg=None):
        req = TrialStartRequest()

        if trial_cfg:
            req.config.content = trial_cfg.SerializeToString()

        rep = self.stub.Start(req)

        observation = DecodeObservationData(
            actor_class, rep.observation.data)

        return ClientTrial(
            self, rep, self.settings, actor_class, rep.actor_counts,
            observation)


class Connection(_Connection_impl):
    def __init__(self, settings, endpoint):
        channel = grpc.insecure_channel(endpoint)
        stub = TrialStub(channel)
        super().__init__(stub, settings)



from cogment.hooks_service import HooksService, TrialHooks
from cogment.agent_service import AgentService, Agent
from cogment.env_service import EnvService, Environment
from cogment.utils import list_versions

from cogment.api.hooks_pb2_grpc import add_TrialHooksServicer_to_server
from cogment.api.environment_pb2_grpc import add_EnvironmentServicer_to_server
from cogment.api.agent_pb2_grpc import add_AgentServicer_to_server

from cogment.api.environment_pb2 import _ENVIRONMENT as env_descriptor
from cogment.api.agent_pb2 import _AGENT as agent_descriptor
from cogment.api.hooks_pb2 import _TRIALHOOKS as hooks_descriptor

from cogment.errors import ConfigError

from grpc_reflection.v1alpha import reflection
from concurrent.futures import ThreadPoolExecutor
import grpc
import os
import signal
from threading import Event

from distutils.util import strtobool

ENABLE_REFLECTION_VAR_NAME = 'COGMENT_GRPC_REFLECTION'
DEFAULT_PORT = 9000
MAX_WORKERS = 10


# A Grpc endpoint serving a cogment service
class GrpcServer:
    def __init__(self, service_type, settings, port=DEFAULT_PORT):
        print("Versions:")
        for v in list_versions(service_type).versions:
            print(f'  {v.name}: {v.version}')

        self._port = port
        self._grpc_server = grpc.server(ThreadPoolExecutor(
          max_workers=MAX_WORKERS))

        # Register service
        if issubclass(service_type, Agent):
            self._service_type = agent_descriptor
            add_AgentServicer_to_server(
              AgentService(service_type, settings), self._grpc_server)
        elif issubclass(service_type, Environment):
            self._service_type = env_descriptor
            add_EnvironmentServicer_to_server(
              EnvService(service_type, settings), self._grpc_server)
        elif issubclass(service_type, TrialHooks):
            self._service_type = hooks_descriptor
            add_TrialHooksServicer_to_server(
              HooksService(service_type, settings), self._grpc_server)
        else:
            raise ConfigError('Invalid service type')

        # Enable grpc reflection if requested
        if strtobool(os.getenv(ENABLE_REFLECTION_VAR_NAME, 'false')):
            SERVICE_NAMES = (
                self._service_type.full_name,
                reflection.SERVICE_NAME,
            )
            reflection.enable_server_reflection(
              SERVICE_NAMES, self._grpc_server)

        self._grpc_server.add_insecure_port(f'[::]:{port}')

    def serve(self):
        self.start()
        print("start serving...")
        stop_serving_event = Event()

        def finished(a, b):
            stop_serving_event.set()

        for sig in ('TERM', 'HUP', 'INT'):
            signal.signal(getattr(signal, 'SIG'+sig), finished)

        stop_serving_event.wait()
        print("done serving...")

    def start(self):
        self._grpc_server.start()
        print(f"{self._service_type.full_name} service"
              f" listening on port {self._port}")

    def stop(self):
        self._grpc_server.stop(0)
